// This file is part of the brownmove simulation package
// 
// (C) 2003 - 2009 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove


#ifndef CUP_H
#define CUP_H

#include <vector>
#include <string>
#include <list>
#include <map>
#include <iostream>

#include <stdio.h>

#include "types.h"
#include "vectormath.h"
#include "types.h"
#include "constants.h"


using namespace std;

namespace brown{
	
	class Molecule;
	class Wall;
	
	//! @author jni
	//
	//! Representing the "physical" container of the molecules as well as 
	//! the implementation-side container of all active class instances
	class Cup {
		public: 
		//! Constructor
		Cup();
		//! Destructor
		~Cup();
		//! Reset the Cup to an empty state
		void initialize();
		//! Set the integration timestep
		void setTimestep(double);
		
		//! Tell the Cup that we want a periodic setup
		void setUnitCellSize(const double dx, const double dy, const double dz);

		
		//! Here is the main engine: Perform one timestep: calculate forces, apply corrections, and propagate the molecules
		void doOneStep(const timestamp &it);

		
		//! describe yourself in a string (currently unused)
		string describe() const;
		
		//! dump current  molecule positions in a browndef format
		void dumpMolecules();
		
		//! @name molecule handlers : This is what you can do with a singel molecule in the cup
		//!@{
		//
		//! Add a molecule to the list.
		//! To be called upon first filling the cup
		//! To be called from one of the injectors if a new molecule comes in. 
		void addMolecule(const Molecule &newMolecule);
		
		//! Return the current value of the molecule-ID counter and increment it.
		//! Used to assign a unique ID to each molecule
		long int getNextMoleculeID();
		
		//! Removes a specific molecule from the cup. To be called by an acceptor.
		//! First version: by position (index) in the molecule list
		void removeMolecule(int pos);
		
		//! Removes a specific molecule from the cup.
		//! Second version: by explicit reference
		void removeMolecule(Molecule* aMolecule);
		
		//! Add a molecule to the list of molecules to be deleted at the end of the current timestep
		void doomMolecule(Molecule *m);
		
		//! Get rid of doomed molecules
		void removeDoomedMols();
		//!@}
		
		//! @name Wall handlers : adding and clearing walls for given setup
		//! @{
		//
		//! Add a new wall to the list of walls of a cup
		void addWall(const Wall& newWall);
		
		//! Remove all walls from cup.
		//! No need for removal of single wall, as the setup shouldn't change during a simulation(?)
		void clearWalls();
		//!@}
		
		
		//! @name Diagnostics : these can be called from the hooks to print intermediate states of the model
		//! @{
		//
		//! Returns the current number of molecules
		long int countMolecules() { return Active_Molecules.size(); }
		//! Returns the current number of molecules of class mclass
		long int countMolecules(string mclass);
		//! Lists molecules in the active_molecule list and their properties
		void listMolecules();

		//! Returns the density of the reservoir with the label resName
		const double getDensity(string resName);
		
		// Init Output of molecule trajectories (obsolete)
		// void initMolPositionsToFile();
		// tells the molecules to append their full position to their trajectory (obsolete)
		// void molPositionsToFile(double absTime);
		
		// Tell the molecules to append their ID and full pos to a 
		// snapshot file; V1: create new file on disk
		// void takeMolSnapshot(long int count, double absTime, char mode[]);

		//! Tell the molecules to append their ID and full pos to a 
		//! snapshot file; V2: append to existing file, i.e., pipe
		void takeMolSnapshot(FILE *fp, double absTime);
		//! Filtered snapshot: only the proteins occuring in auswahl are 
		//! listed in the output
		void takeMolSnapshot_filtered(FILE *fp, double absTime, list<string> auswahl);
		

		//! Returns the current number of walls
		int countWalls() { return My_Walls.size(); }
		//! lists fullPositions of all wall in the Cup
		void listWalls();
		
		//! Returns full force on the first copy of the molecule with the given label.
		fullForce getProbeForce(string probeLabel);
		
		// Print forces on all particles currently in the Cup
		// void printForces();
		//! @}
		
		
		
		//! @name Handling the vdW params
		//! @{
		//
		//! Initialize array of given dimension to hold dim^2 vdW parameter sets
		void initializeVDWArray(long int dim);
		//! Initialize array elements [i,k] and [k,i] with the given values (symmetric in the index); fails if i,k > maxDim
		void setVDWParam(const long int i, const long int k, const double C_12, const double C_6, const double C_3, const double r_0, const double dr, const double rmin, const double rmax);
		//! initialize array element i,k from a given struct P; fails if i,k > maxDim
		void setVDWParam(const long int i, const long int k, const vdWParameter &P) {
			setVDWParam(i, k, P.C_12, P.C_6, P.C_3, P.R_0, P.DR, P.Rmin, P.Rmax);
		}
		//! @}
		
		
		//! @name Handling Reservoirs
		//! @{
		//
		//! Add a new reservoir with name, initial concentration, and inverse volume
		void addReservoir(std::string name, double conc, double ivol);
		//! @}
		
		
		// ###################
		private: 
		
		//! @name loops : Here are the parts of the main engine, which can only be accessed from the public doOneStep()
		//! @{
		//
		//! Tell the molecules to tell their Gestalt objects to initialize the random forces.
		void randomForcesOnMolecules();
		
		//! Have all molecules interact with each wall in turn.
		void interactMoleculeswithWalls();
		
		//! Have all molecules interact with all other molecules
		void interactMoleculeswithMolecules();
		
		//! Calculate and apply Langevin corrections (= finite damping of the veocities) to the forces
		void correctDamping();
		
		//! Have all molecules interact with all others through HI: gets external forces, calculates random forces
		void interactHi();

		//! tell all molecules to move
		void tellMoleculesToMove();
		
		//! Tell the walls to insert new molecules through their injectors
		void insertNewMolecules(const timestamp &it);
		//@}
		
		//! Immobilize all molecules in the Cup for n timesteps
		void immobilizeMolecules(long int n);

		
		/** active molecules in the container 
		The simplest implementation is a list.
		Future Versions of "cup" can contain a more sophisticated data structure
		to facilitate finding neighboring molecules etc.
		*/
		list<Molecule*> Active_Molecules;
		list<Molecule*> Doomed_Molecules;
		
		//! Vector of the walls in the cup
		vector<Wall*> My_Walls;
				
		//! Current Simulation Time
		timestamp Current_Time;
		
		//! Counter for next unique molecule id
		long int Next_Molecule_Id;
		
		//! Simulation timestep
		double timestep;
		
		//! Pointer for the vdW-Parameters-Array
		vdWParameter **Vdw_Array;
		//! Max. index in the vdw-array
		long int Vdw_Array_Max_Index;
		
		//! Map of reservoirs 
		map<std::string, struct reservoir> resMap;
		
		//! Size of a periodic simulation box, (-1, -1, -1) if not periodic
		doublevec unitCellSize;
		//! True, if quasi periodic in x, y, or z direction, resp.
		bool isPeriodic[3];
	};
} // end namespace

#endif
