// This file is part of the brownmove simulation package
// 
// (C) 2003 - 2009 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove


#ifndef INTERACTINGBODY_H
#define INTERACTINGBODY_H

#include <vector>

#include "gestalt.h"
#include "bondshape.h"

#include "types.h"
#include "vectormath.h"


namespace brown{

	//! @author jni,tg
	
	//! Base class for all the stuff in the simulation: molecules, walls, and obstacles
	class InteractingBody {
		// friend class MoleculeInjector;
		
		public:
		//! Constructor
		InteractingBody(string name, long int num);
		
		//! Copy constructor, does not change/set the unique(!) id. 
		//! This is done in the derived molecule copy constructor.
		//! Walls copy full id, they do not care about the id.mnumber.
		InteractingBody(const InteractingBody &ib);
		
		//! Destructor
		virtual ~InteractingBody();
		
		
		//! @name Various query methods
		//! @{
		const fullPosition &getFullPosition() 		const {return myPosition;};
		const position     &getPosition()			const {return myPosition.pos;};
		const rotation     &getRotation()     		const {return myPosition.rot;};
		//! @}

		
		//! Calculate and return (a ref to the) center of mass determined
		//! from the vdW-Spheres, for which a constant density is assumed. 
		//! If no vdW-spheres are defined, the position of the Gestalt (=its
		//! origin) is returned.
		const position &getCenterOfMass();
		
		/** Putting the object on a given point
		(to be exact: put its origin on that point, and rotating it to the given angle)*/
		void teleport(const fullPosition &p);
		
		
		//! this actually moves the thing.
		void move(const position &lateralMove, const rotation &rotationMove);
		//! move the thing by a fullPosition (wrapper for move(dx, dphi))
		inline void move(const fullPosition &pd) { move(pd.pos, pd.rot);}
		
		//! add an existing gestalt to the list
		void addGestalt(Gestalt *g){
			Gestalten.push_back(*g);
		}
		
		std::vector<Gestalt> *getGestalten() {return &Gestalten;};
		
		//! returns the number of Gestalt objects
		int size() { return(Gestalten.size()); }
		
		//! Return pointer to Gestalt identified by label
		Gestalt *getGestalt(std::string label);
		
		//! @name Handling Bonds
		//@{
		//! add bondParameter to the vector, returns index in the vector
		long int addBondParameter(const double &R_0, const double &k, const std::string &label);
		//! return pointer to the bondVector
		std::vector<bondParameter> *getBondVector() { return(&bondVector); }
		//@}
	
		//! Set pointer to the vdW parameter array that lives globally accessible in the cup
		void setvdWParameterArray(vdWParameter **Vdw_Array);
		
		
		//! Read molecule id (number + string label)
		moleculeID getid() const {return id;};

		//! Set the mclass (name) part of the molecule id
		void setName(std::string theName) { id.mclass = theName; }
		
		
		//! Write out the internal structure in a browndef-Style for re-reading
		void dumpStructure();
		
		
		protected:
		//! unique identifier, fixed when a molecule is inserted into the simulation volume
		moleculeID id;

		//! Full position consisting of the absolute position of its origin (= coordinate system origin for calculation of shapes)
		//! and the rotation (the combined effect of all occured rotations since its creation).
		fullPosition myPosition;

		//! The absolute position of the center of mass, may be different from myPosition
		position myCM;

		
		//! Vector for the gestalt objects, which contain the sub-parts of the molecule (the "real" things)
		std::vector<Gestalt> Gestalten;
		
		//! Pointer to the vector of bond-parameters
		std::vector<bondParameter> bondVector;

		private:
		// we're so sorry, but there's no privacy around here
	};
}
#endif /* INTERACTINGBODY_H */
