// This file is part of the brownmove simulation package
// 
// (C) 2003 - 2009 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove


#ifndef MOLECULEACCEPTOR_H
#define MOLECULEACCEPTOR_H

#include <vector>
#include <map>

#include "molecule.h"
#include "moleculehandler.h"
#include "cup.h"

#include "types.h"
#include "vectormath.h"

namespace brown{
	/** An Interface for classes that can accept molecules 
	 *  e.g. from a wall that swallows them.
	 *  
	 *  @author jni, tg
	 */
	class MoleculeAcceptor : public MoleculeHandler {
		public: 
		//! Default constructor with a label
		MoleculeAcceptor(string newLabel);
		//! Copy constructor
		MoleculeAcceptor(const MoleculeAcceptor &other);
		//! Destructor
		~MoleculeAcceptor(){};

		//! Return descriptive string
		std::string describe();
		
		
		//! Struct to hold the molecule id strings, a pointer to the density and an inverse
		//! Volume, so that mols taken out can be inserted somewhere else.
		//! With a negative inverse volume, particles are counted, otherwise, a density is used.
		//! If the pointer to the vector<fullPosition> is set, the position of the accepted
		//! molecule is saved there to be reused at insertion -- makes sense only with a negative invVol.
		struct acceptTemplate {
			string idString;
			string resName;
			double *concentration;
			double invVol;
			vector<fullPosition> *accPos;
			int *theInternal;
			int resCond;
			long int immobilizeT;
		};
		
		
		//! Take this molecule and check whether it is inside the bounds defined
		//! by the mask. If so, return true. You then have to remove it from the cup, the acceptor
		//! does not delete it itself.
		//! To save time the already computed distance vector between the molecule and the wall
		//! is passed to accept.
		bool accept(Molecule *m, const position &dr);
		
		//! set the pointer to a mask function 
		void setMask(acceptPtr &ap) {
			myMask = ap;
		}
		
		//! Add a molecule id and reservoir
		void addMolecule(const string id, double *concentration, const double &invVol, vector<fullPosition> *accPos, int *theInternal, int resCond, long int immobilizeT);
		void addMolecule(std::string molName, std::string resName);
		
		//! Resolve names of the reservoirs
		void resolveMolNames(std::map<std::string, Molecule*> *molMap);
		void resolveResNames(std::map<std::string, struct reservoir> *resMap);

		
		private:
		acceptPtr myMask;
		// vector of molecules and their reservoirs
		vector<acceptTemplate> reservoirs;
	};
}
#endif
