// This file is part of the brownmove simulation package
// 
// (C) 2003 - 2009 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove



//! @todo optimize these: use Practical C++ Programming p. 309 ff.
//! instead of indexing the arrays, use pointers and pointer increments

//! this is not object orientation:
//! position should be a class with the method move.
//! for simple types like the position is, we don't do that
//! for optimization reasons.

#ifndef VECTORMATH_H
#define VECTORMATH_H

#include <math.h>

#include "types.h"


namespace brown {
	// ----- operations for doublevecs ------
	inline  void doublevec_add(position& thePosition, const position& theDelta){
		thePosition[0]+=theDelta[0];
		thePosition[1]+=theDelta[1];
		thePosition[2]+=theDelta[2];
	}

	inline  void doublevec_sub(position& thePosition, const position& theDelta){
		thePosition[0]-=theDelta[0];
		thePosition[1]-=theDelta[1];
		thePosition[2]-=theDelta[2];
	}
	
	inline  void doublevec_mult(position& thePosition, const double& factor){
		thePosition[0]*=factor;
		thePosition[1]*=factor;
		thePosition[2]*=factor;
	}
	
	//! changes the argument position.
	inline  void positionMove(position& thePosition, const position& theDelta){
		thePosition[0]+=theDelta[0];
		thePosition[1]+=theDelta[1];
		thePosition[2]+=theDelta[2];
	}
	
	
	inline  void positionMinus(position& thePosition, const position& theDelta){
		thePosition[0]-=theDelta[0];
		thePosition[1]-=theDelta[1];
		thePosition[2]-=theDelta[2];
	}
		
	inline void doublevec_set(doublevec &dv, double d1){
		dv[0]=d1; dv[1]=d1; dv[2]=d1; 
	}
	
	inline void doublevec_set(doublevec &dv, double d1, double d2, double d3){
		dv[0]=d1; dv[1]=d2; dv[2]=d3; 
	}
	
	inline void doublevec_set(doublevec &dv, const doublevec &other){
		dv[0]=other[0]; dv[1]=other[1]; dv[2]=other[2]; 
	}
	
	inline  bool doublevec_equals(const doublevec &p1, const doublevec &p2){
		return((p1[0] == p2[0]) && (p1[1] == p2[1]) && (p1[2] == p2[2]));
	}
	
	inline  bool doublevec_equals_num(const doublevec &p1, const doublevec &p2, const double &eps){
		return((fabs(p1[0]-p2[0]) <= eps) && (fabs(p1[1]-p2[1]) <= eps) && (fabs(p1[2]-p2[2]) <= eps));
	}
	
	inline double doublevec_length(const doublevec &v) {
		return(sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]));
	}
	
	
	inline double doublevec_scalprod(const doublevec &v1, const doublevec &v2) {
		return(v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2]);
	}
	
	inline void doublevec_kreuzprod(doublevec &erg, const doublevec &v1, const doublevec &v2)
	{
		erg[0] = v1[1]*v2[2] - v1[2]*v2[1];
		erg[1] = v1[2]*v2[0] - v1[0]*v2[2];
		erg[2] = v1[0]*v2[1] - v1[1]*v2[0];
	}
	
	inline void doublevec_rotate(position &p, const rotation &r) {
		position tmp1, tmp2;
		double cp = cos(r.phi);
		double sp = doublevec_scalprod(r.n, p);
		doublevec_set(tmp1, r.n);
		doublevec_mult(tmp1, sp*(1.0 - cp));
		doublevec_kreuzprod(tmp2, r.n, p);
		doublevec_mult(tmp2, sin(r.phi));
		doublevec_mult(p, cp);
		doublevec_add(p, tmp1);
		doublevec_add(p, tmp2);
	}
	
	
	// ---- operations for quaternions -----
	inline void quaternion_set(quaternion &q, const double &d) {
		q[0] = d; q[1] = d; q[2] = d; q[3] = d;
	}
	
	inline void quaternion_set(quaternion &q, const double &d0, const double &d1, const double &d2, const double &d3) {
		q[0] = d0; q[1] = d1; q[2] = d2; q[3] = d3;
	}
	
	inline void quaternion_set(quaternion &q, const quaternion &o) {
		q[0] = o[0]; q[1] = o[1]; q[2] = o[2]; q[3] = o[3];
	}
	
	inline bool quaternion_equals(const quaternion &p1, const quaternion &p2){
		return((p1[0] == p2[0]) && (p1[1] == p2[1]) && (p1[2] == p2[2]) && (p1[3] == p2[3]));
	}
	
	// erg = q1 * q2;
	inline void quaternion_mult(quaternion &erg, const quaternion &q1, const quaternion &q2) {
		erg[0] = q1[0]*q2[0] - q1[1]*q2[1] - q1[2]*q2[2] - q1[3]*q2[3];
		erg[1] = q1[0]*q2[1] + q1[1]*q2[0] + q1[2]*q2[3] - q1[3]*q2[2];
		erg[2] = q1[0]*q2[2] + q1[2]*q2[0] - q1[1]*q2[3] + q1[3]*q2[1];
		erg[3] = q1[0]*q2[3] + q1[3]*q2[0] + q1[1]*q2[2] - q1[2]*q2[1];
	}
	
	inline void quaternion2rotation(rotation &rot, const quaternion &q)
	{
		double isinphi;
		if (fabs(q[0]) == 1.0) {
			rot.phi = 0.0;
			rot.n[0] = 1.0; rot.n[1] = 0.0; rot.n[2] = 0.0;   // could also have chosen another axis...
		} else {
			rot.phi = 2.0 * acos(q[0]);
			isinphi = 1.0/sin(0.5 * rot.phi);
			rot.n[0] = isinphi * q[1];
			rot.n[1] = isinphi * q[2];
			rot.n[2] = isinphi * q[3];
		}
	}
	
	inline void rotation2quaternion(quaternion &q, const rotation &rot)
	{
		double sinphi = sin(0.5 * rot.phi);
		q[0] = cos(0.5 * rot.phi);
		q[1] = rot.n[0]*sinphi; q[2] = rot.n[1]*sinphi; q[3] = rot.n[2]*sinphi;
	}
	
	inline void direction2rotation(rotation &r, const direction &d)
	{
		r.phi = sqrt(d[0]*d[0] + d[1]*d[1] + d[2]*d[2]);
		if(fabs(r.phi) <= 1.0e-10)
		{
			r.n[0] = 1.0;
			r.n[1] = 0.0;
			r.n[2] = 0.0;
		} else {
			double ilang = 1.0/r.phi;
			r.n[0] = d[0]*ilang;
			r.n[1] = d[1]*ilang;
			r.n[2] = d[2]*ilang;
		}
	}
	
	inline void rotation2direction(direction &d, const rotation &r)
	{
		d[0] = r.phi * r.n[0];
		d[1] = r.phi * r.n[1];
		d[2] = r.phi * r.n[2];
	}
	
	// -------- operations for rotations ------------
	inline void rotation_set(rotation &r1, const double &phi, const double &n0, const double &n1, const double &n2) {
		r1.phi = phi;
		doublevec_set(r1.n, n0, n1, n2);
	}
	
	inline void rotation_set(rotation &r1, const rotation &r2) {
		r1.phi = r2.phi;
		doublevec_set(r1.n, r2.n);
	}
	
	// Comparison
	inline  bool rotation_equals(const rotation &r1, const rotation &r2){
		return((r1.phi == r2.phi) && doublevec_equals(r1.n, r2.n));
	}
	
	// Comparison with tolerance eps
	inline  bool rotation_equals_num(const rotation &r1, const rotation &r2, const double &eps){
		return((fabs(r1.phi - r2.phi) <= eps) && doublevec_equals_num(r1.n, r2.n, eps));
	}
	
	
	// ----- operations for tripleMatrices -----
	inline void trMat_set(tripleMatrix &erg, const double &a) {
		erg[0][0] = a;   erg[0][1] = 0.0; erg[0][2] = 0.0;
		erg[1][0] = 0.0; erg[1][1] = a;   erg[1][2] = 0.0;
		erg[2][0] = 0.0; erg[2][1] = 0.0; erg[2][2] = a;
	}
	
	inline void trMat_set(tripleMatrix &erg, const tripleMatrix &a) {
		erg[0][0] = a[0][0];   erg[0][1] = a[0][1]; erg[0][2] = a[0][2];
		erg[1][0] = a[1][0];   erg[1][1] = a[1][1]; erg[1][2] = a[1][2];
		erg[2][0] = a[2][0];   erg[2][1] = a[2][1]; erg[2][2] = a[2][2];
	}
	
	inline void trMat_add(tripleMatrix &mat, const tripleMatrix &a) {
		mat[0][0] += a[0][0]; mat[1][0] += a[1][0]; mat[2][0] += a[2][0];
		mat[0][1] += a[0][1]; mat[1][1] += a[1][1]; mat[2][1] += a[2][1];
		mat[0][2] += a[0][2]; mat[1][2] += a[1][2]; mat[2][2] += a[2][2];
	}

	inline void trMat_sub(tripleMatrix &mat, const tripleMatrix &a) {
		mat[0][0] -= a[0][0]; mat[1][0] -= a[1][0]; mat[2][0] -= a[2][0];
		mat[0][1] -= a[0][1]; mat[1][1] -= a[1][1]; mat[2][1] -= a[2][1];
		mat[0][2] -= a[0][2]; mat[1][2] -= a[1][2]; mat[2][2] -= a[2][2];
	}

	// multiply: erg = mat * vec
	inline void trMat_mult(doublevec &erg, const tripleMatrix &mat, const doublevec &vec) {
		erg[0] = mat[0][0]*vec[0] + mat[0][1]*vec[1] + mat[0][2]*vec[2];
		erg[1] = mat[1][0]*vec[0] + mat[1][1]*vec[1] + mat[1][2]*vec[2];
		erg[2] = mat[2][0]*vec[0] + mat[2][1]*vec[1] + mat[2][2]*vec[2];
	}

	inline void trMat_mult(tripleMatrix &mat, double a) {
		mat[0][0]*= a; mat[1][0]*= a; mat[2][0]*= a;
		mat[0][1]*= a; mat[1][1]*= a; mat[2][1]*= a;
		mat[0][2]*= a; mat[1][2]*= a; mat[2][2]*= a;
	}

	// ----- operations for diffusionMatrices -----

	// multiply 6x6 diffusion matrix with 6D (force,torque) vector 
	inline void transform(doublevec &ergforce, doublevec &ergtorque, const diffusionMatrix &matr, const doublevec &force, const doublevec &torque) {
		ergforce[0] =  matr[0][0] * force[0]  + matr[0][1] * force[1]  + matr[0][2] * force[2] + matr[0][3] * torque[0] + matr[0][4] * torque[1] + matr[0][5] * torque[2];
		ergforce[1] =  matr[1][0] * force[0]  + matr[1][1] * force[1]  + matr[1][2] * force[2] + matr[1][3] * torque[0] + matr[1][4] * torque[1] + matr[1][5] * torque[2];
		ergforce[2] =  matr[2][0] * force[0]  + matr[2][1] * force[1]  + matr[2][2] * force[2] + matr[2][3] * torque[0] + matr[2][4] * torque[1] + matr[2][5] * torque[2];
		ergtorque[0] = matr[3][0] * force[0]  + matr[3][1] * force[1]  + matr[3][2] * force[2] + matr[3][3] * torque[0] + matr[3][4] * torque[1] + matr[3][5] * torque[2];
		ergtorque[1] = matr[4][0] * force[0]  + matr[4][1] * force[1]  + matr[4][2] * force[2] + matr[4][3] * torque[0] + matr[4][4] * torque[1] + matr[4][5] * torque[2];
		ergtorque[2] = matr[5][0] * force[0]  + matr[5][1] * force[1]  + matr[5][2] * force[2] + matr[5][3] * torque[0] + matr[5][4] * torque[1] + matr[5][5] * torque[2];
	}
	
	// multiply transposed 6x6 diffusion matrix with 6D (force,torque) vector
	inline void transform_T(doublevec &ergforce, doublevec &ergtorque, const diffusionMatrix &matr, const doublevec &force, const doublevec &torque) {
		ergforce[0] =  matr[0][0] * force[0]  + matr[1][0] * force[1]  + matr[2][0] * force[2] + matr[3][0] * torque[0] + matr[4][0] * torque[1] + matr[5][0] * torque[2];
		ergforce[1] =  matr[0][1] * force[0]  + matr[1][1] * force[1]  + matr[2][1] * force[2] + matr[3][1] * torque[0] + matr[4][1] * torque[1] + matr[5][1] * torque[2];
		ergforce[2] =  matr[0][2] * force[0]  + matr[1][2] * force[1]  + matr[2][2] * force[2] + matr[3][2] * torque[0] + matr[4][2] * torque[1] + matr[5][2] * torque[2];
		ergtorque[0] = matr[0][3] * force[0]  + matr[1][3] * force[1]  + matr[2][3] * force[2] + matr[3][3] * torque[0] + matr[4][3] * torque[1] + matr[5][3] * torque[2];
		ergtorque[1] = matr[0][4] * force[0]  + matr[1][4] * force[1]  + matr[2][4] * force[2] + matr[3][4] * torque[0] + matr[4][4] * torque[1] + matr[5][4] * torque[2];
		ergtorque[2] = matr[0][5] * force[0]  + matr[1][5] * force[1]  + matr[2][5] * force[2] + matr[3][5] * torque[0] + matr[4][5] * torque[1] + matr[5][5] * torque[2];
	}
	
	
	// fill diagonal of a diffusionmatrix with constant a, initialize off-diagonals with 0
	inline void makeDiagonal(diffusionMatrix &m, const double a) {
	    for(int i = 0; i<6; ++i) {
	        m[i][i] = a;
	    }
	    for(int i=0; i<6; ++i) {
	        for(int j=i+1; j<6;++j) {
	            m[i][j] = 0.0;
	            m[j][i] = 0.0;
	        }
	    }
	}
}

#endif
