// This file is part of the brownmove simulation package
// 
// (C) 2003 - 2009 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove


#include "hbondshape.h"

namespace brown{
	// access to global variables
	extern double HBOND_DEPTH;
	extern double HBOND_LENGTH;
	extern double HBOND_WIDTH;
	extern double HBOND_CUTOFF;

	//! (Empty) Constructor
	HBondShape::HBondShape(){
	}
	
	//! Copy constructor
	HBondShape::HBondShape(const HBondShape &other){
		hooks = other.hooks;
		actual_hooks = other.actual_hooks;
	}
	
	//! Destructor
	HBondShape::~HBondShape(){
		hooks.clear();
		actual_hooks.clear();
	}
	
	
	std::string HBondShape::describe(){
		ostringstream ausgabe("", ios::out);
		int i=0;
		
		ausgabe << "actual charge positions";
		vector<hbondPos>::iterator mlc;
		for(mlc = actual_hooks.begin(); mlc != actual_hooks.end(); ++mlc, i++){
			ausgabe << " : "<<(*mlc).pos[0]<<" "<<(*mlc).pos[1]<<" "<<(*mlc).pos[2]<<" ("<<(*mlc).bond[0]<<" "<<(*mlc).bond[1]<<" "<<(*mlc).bond[2]<<")" ; 
		}
		ausgabe <<" :";
		for(mlc = hooks.begin(); mlc != hooks.end(); ++mlc, i++){
			ausgabe << " : "<<(*mlc).pos[0]<<" "<<(*mlc).pos[1]<<" "<<(*mlc).pos[2]<<" ("<<(*mlc).bond[0]<<" "<<(*mlc).bond[1]<<" "<<(*mlc).bond[2]<<")" ; 
		}
		return(ausgabe.str());
	}
	
	
	void HBondShape::recalculate(const fullPosition &p) 
	{
		vector<hbondPos>::iterator act = actual_hooks.begin();
		for(vector<hbondPos>::iterator mlc = hooks.begin(); mlc != hooks.end(); ++mlc, ++act)
		{
			doublevec_set((*act).pos, (*mlc).pos);
			doublevec_rotate((*act).pos, p.rot);
			positionMove((*act).pos, p.pos);
			doublevec_set((*act).dr, (*act).pos);
			doublevec_sub((*act).dr, p.pos);
			
			doublevec_set((*act).bond, (*mlc).bond);
			doublevec_rotate((*act).bond, p.rot);
		}
	}
	
	
	void HBondShape::addHook(hbondPos hp, const position &moleculeCenter){
		doublevec_set(hp.dr, hp.pos);
		doublevec_sub(hp.dr, moleculeCenter);
		hooks.push_back(hp);
		actual_hooks.push_back(hp);
	}
	
	
	void HBondShape::calculateImpact(fullForce &otherForce, fullForce &myForce, HBondShape *hs, const bool isPeriodic[3], const doublevec &cellSize)
	{
		position dr, ri, dr_temp;
		double r12, prefactor, B12, help;
		double myScale, otherScale, winkel;
		force df;
		torque dt;
		
		// interact all hooks with each other -- nothing happens, if there are none
		// Begin with the other's hooks -- we have at least one or we should not be here
		for(vector<hbondPos>::iterator mlc2 = hs->actual_hooks.begin(); mlc2 != hs->actual_hooks.end(); ++mlc2)
		{
			otherScale = (*mlc2).giveOrTake;	// donor has scale > 0.0
			
			// cerr <<"*** HBondShape::calculateImpact  mlc2:   "<<"  " << (*mlc2).pos[0] << "  " << (*mlc2).pos[1] << "  " << (*mlc2).pos[2] << endl;

			// Now my own hooks
			for(vector<hbondPos>::iterator mlc1 = actual_hooks.begin(); mlc1 != actual_hooks.end(); ++mlc1)
			{
				// check for donor-acceptor pair
				myScale = (*mlc1).giveOrTake;
				if(myScale * otherScale < 0.0)
				{
					
					// cerr <<"*** HBondShape::calculateImpact  mlc1: "<< "   " << (*mlc1).pos[0] << "  " << (*mlc1).pos[1] << "  " << (*mlc1).pos[2] << endl;
					
					// calculate relative distance vector donor-h -> acceptor
					if(myScale > 0.0) {		// I'm the donor
						doublevec_set(dr, (*mlc1).pos);
						doublevec_sub(dr, (*mlc2).pos);
					} else {
						doublevec_set(dr, (*mlc2).pos);
						doublevec_sub(dr, (*mlc1).pos);
					}
					
					double dx = 0.0;
					double dy = 0.0;
					double dz = 0.0;
					int ix = 2, iy = 2, iz = 2;
					
					// if periodic, start with displaced hooks
					if(isPeriodic[0]) {
						dx = -cellSize[0];
						ix = 0;
					}
					if(isPeriodic[1]) {
						dy = -cellSize[1];
						iy = 0;
					}
					if(isPeriodic[2]) {
						dz = -cellSize[2];
						iz = 0;
					}
					
					while(true)
					{
						// this is for the (quasi) periodic setup
						dr_temp[0] = dr[0] + dx;
						dr_temp[1] = dr[1] + dy;
						dr_temp[2] = dr[2] + dz;
						
						r12 = doublevec_length(dr_temp);
						// check if distance between hooks is shorter than cut-off
						if(r12 <= HBOND_CUTOFF) 
						{
							// angular term: scalar product between the two vectors
							if(myScale > 0.0) {
								winkel = doublevec_scalprod(dr_temp, (*mlc1).bond);
							} else {
								winkel = doublevec_scalprod(dr_temp, (*mlc2).bond);
							}
							winkel /= r12;	// correct for length of dr_temp
							
							// interaction potential: inverted Gaussian
							help = r12 - HBOND_LENGTH;
							prefactor = (HBOND_DEPTH / (HBOND_WIDTH*HBOND_WIDTH)) * myScale * otherScale * help * exp(-help*help/(2.0*HBOND_WIDTH*HBOND_WIDTH));  // let the compiler calculate all the constants
							prefactor /= r12;		// for normalization of dr_temp
							
							// multiply both contributions
							prefactor *= winkel;
							
							// cerr <<"*** HBondShape::calculateImpact: r12 = " << r12 << "  prefactor = " << prefactor << endl;
							
							// vectorial force between donor and acceptor
							doublevec_set(df, dr_temp);				// for the direction
							doublevec_mult(df, prefactor);
							
							// add force to previous results
							doublevec_add(myForce.f, df);
							doublevec_sub(otherForce.f, df);
							
							// calculate and add torques
							doublevec_kreuzprod(dt, (*mlc1).dr, df);
							doublevec_add(myForce.t, dt);
							doublevec_kreuzprod(dt, (*mlc2).dr, df);
							doublevec_sub(otherForce.t, dt);
						}
						
						
						// for the quasi periodic boundaries: displace temp. hook positions and redo above evaluation
						if(ix < 2) {
							ix++;
							dx += cellSize[0];
						}
						else
						{
							if(isPeriodic[0]) {
								ix = 0;
								dx = -cellSize[0];
							}
							
							if(iy < 2) {
								iy++;
								dy += cellSize[1];
							}
							else
							{
								if(isPeriodic[1]) {
									iy = 0;
									dy = -cellSize[1];
								}
								
								if(iz < 2) {
									iz++;
									dz += cellSize[2];
								}
								else
								{
									break;
								}
							}
						}
					}
				}
			}
		}
	}

	
	
	void HBondShape::dumpStructure() {
		std::cout << "\n\t\tHBondShape myHHooks" << std::endl;
		
		for(std::vector<hbondPos>::iterator it = hooks.begin(); it != hooks.end(); it++) {
			if((*it).giveOrTake > 0.0) {
				std::cout << "\t\t\tdonor   "<< (*it).pos[0]-(*it).bond[0]<<" "<<(*it).pos[1]-(*it).bond[1]<<" "<<(*it).pos[2]-(*it).bond[2]<<"   "<< (*it).pos[0]<<" "<<(*it).pos[1]<<" "<<(*it).pos[2]<<"   "<<(*it).giveOrTake<< std::endl;
			}
			else {
				std::cout << "\t\t\tacceptor   "<< (*it).pos[0]<<" "<<(*it).pos[1]<<" "<<(*it).pos[2]<<"   "<<(-(*it).giveOrTake)<< std::endl;
			}
		}
		
		std::cout << "\t\tEnd" << std::endl;
	}

	
}
