// This file is part of the brownmove simulation package
// 
// (C) 2003 - 2009 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove


#include "moleculeacceptor.h"

namespace brown {
	// simple "allways-true" function to be used as a default mask
	bool defaultAcceptorMask(const double &y, const double &z) {
		//! @todo The default mask should use the mask dimensions defined in moleculeHandler
		return(true);
	}
	

	MoleculeAcceptor::MoleculeAcceptor(string newLabel) : MoleculeHandler(newLabel) {
		myMask.mask = defaultAcceptorMask;
		reservoirs.clear();
	}

	MoleculeAcceptor::MoleculeAcceptor(const MoleculeAcceptor &other) : MoleculeHandler(other) {
		myMask.mask = other.myMask.mask;
		reservoirs = other.reservoirs;
	}
	
	
	std::string MoleculeAcceptor::describe() {
		ostringstream ausgabe("", ios::out);
		
		ausgabe << label << ": ";
		
		vector<acceptTemplate>::iterator it;
		for(it = reservoirs.begin(); it != reservoirs.end(); ++it) {
			ausgabe << "  " << it->idString << "@" << it->resName;
		}
		
		return(ausgabe.str());
	}

	
	
	void MoleculeAcceptor::addMolecule(const string id, double *concentration, const double &invVol, vector<fullPosition> *accPos, int *theInternal, int resCond, long int immobilizeT)
	{
		acceptTemplate newRes;
		
		if(theInternal) {
			cout << "*** MoleculeAcceptor::addMolecule: theInternal = " << *theInternal << endl;
		} else {
			cout << "*** MoleculeAcceptor::addMolecule: no internal reaction defined" << endl;
		}

		newRes.idString = id;
		newRes.resName = "";
		newRes.concentration = concentration;
		newRes.invVol = invVol;
		newRes.accPos = accPos;
		newRes.theInternal = theInternal;
		newRes.resCond = resCond;
		newRes.immobilizeT = immobilizeT;
		
		reservoirs.push_back(newRes);
	}
	
	
	void MoleculeAcceptor::addMolecule(std::string molName, std::string resName)
	{
		std::cerr << "MoleculeAcceptor::addMolecule: adding " << molName << " at reservoir " << resName << std::endl;
		
		acceptTemplate newRes;

		newRes.idString = molName;
		newRes.resName = resName;
		newRes.concentration = NULL;
		newRes.invVol = 0.0;
		newRes.accPos = NULL;
		newRes.theInternal = NULL;
		newRes.resCond = 0;
		newRes.immobilizeT = 0L;
		
		reservoirs.push_back(newRes);
	}
	


	void MoleculeAcceptor::resolveResNames(std::map<std::string, struct reservoir> *resMap) 
	{
		std::cerr << "MoleculeInjector::resolveResNames: size(resMap) = " << resMap->size() <<  std::endl;
		
		vector<acceptTemplate>::iterator it;
		for(it = reservoirs.begin(); it != reservoirs.end(); ++it)
		{
			if(resMap->count(it->resName) == 1) {
				it->concentration = &((*resMap)[it->resName].density);
				it->invVol = (*resMap)[it->resName].invVol;
				std::cerr << "MoleculeAcceptor::resolveMolNames: resolved molecule "<< it->idString  << " at " << it->resName << " (" << it->concentration << ")" << std::endl;
				std::cerr << "MoleculeAcceptor::resolveMolNames: concentration = " << *(it->concentration) << std::endl;
			}
			else {
				std::cerr << "MoleculeAcceptor::resolveMolNames: Molecule " << it->idString << " at " << it->resName << " could not be resolved" << std::endl;
				exit(-1);
			}
		}
	}

	
	
	bool MoleculeAcceptor::accept(Molecule *m, const position &dr) 
	{
		//! @todo Have a closer look here, what is still required (=> reactions?)
		double proj_x, proj_y, proj_z;
		
		proj_x = doublevec_scalprod(dr, (*myWallsNormal));
		proj_y = doublevec_scalprod(dr, (*myWallsYAxis));
		proj_z = doublevec_scalprod(dr, (*myWallsZAxis));
		
		// std::cerr << "Moleculeacceptor::accept: dr            = " <<dr[0]<<" "<<dr[1]<<" "<<dr[2]<< std::endl;
		// std::cerr << "Moleculeacceptor::accept: myWallsNormal = " <<(*myWallsNormal)[0]<<" "<<(*myWallsNormal)[1]<<" "<<(*myWallsNormal)[2]<< std::endl;
		// std::cerr << "Moleculeacceptor::accept: myWallsYAxis  = " <<(*myWallsYAxis)[0]<<" "<<(*myWallsYAxis)[1]<<" "<<(*myWallsYAxis)[2]<< std::endl;
		// std::cerr << "Moleculeacceptor::accept: myWallsZAxis  = " <<(*myWallsZAxis)[0]<<" "<<(*myWallsZAxis)[1]<<" "<<(*myWallsZAxis)[2]<< std::endl;
		// std::cerr << "Moleculeacceptor::accept: proj = " << proj_x<< " "<< proj_y <<" "<< proj_z << std::endl;
		
		// std::cerr << "MoleculeAcceptor::accept: proj_y = " << proj_y << "  maskY0 = " << maskY0 << "   maskdY = " << maskdY << std::endl;
		
		// check if mol is inside bounds of mask
		//! @todo Here, the mask from the structure should be used instead of the hard coded rectangle (default mask would be rectangle)
		if((proj_y >= maskY0-maskdY) && (proj_y <= maskY0+maskdY) && (proj_z >= maskZ0-maskdZ) && (proj_z <= maskZ0+maskdZ))
		{
			// std::cerr << "MoleculeAcceptor::accept: mol inside mask -> action" << endl;

			moleculeID mID = m->getid();
			// std::cerr << "MoleculeAcceptor::accept: mol-id = " << mID.mclass << endl;
			
			// iterate through the reservoirs and compare the ids
			for(vector<acceptTemplate>::iterator it = reservoirs.begin(); it != reservoirs.end(); ++it)
			{
				// std::cerr << "MoleculeAcceptor::accept: "<<label<<", mol id = " << mID.mclass << "  templ id = " << (*it).idString << endl;
				if((*it).idString == mID.mclass)
				{
					// std::cerr << "*** MoleculeAcceptor::accept: right molecule "<<(*it).idString<<", checking conditions: " << *((*it).theInternal) <<"<->"<<(*it).resCond<< endl;
					// std::cerr << "*** MoleculeAcceptor::accept: right molecule "<<(*it).idString << endl;
					
					
					bool tryIt = false;
					// check for internal reservoir / reaction -> if none 
					// exists, try to accept the molecule, if it exists, ..
					if((*it).theInternal) {
						// .. check the condition on the internal reservoir
						if((*((*it).theInternal)) == (*it).resCond) {
							tryIt = true;
						}
					} else {
						tryIt = true;
					}
					
					if(tryIt == true)
					{
//						std::cerr << dr[0] << " " << dr[1] << " " << dr[2] << std::endl;
						// cout << "*** MoleculeAcceptor::accept: " << label << " accepting mol " << (*it).idString << "  theInternal = " << *((*it).theInternal) <<  endl;
						
						long int molImmobilized = m->getImmobilized();
						
						if(m->getImmobilized() == 0)
						{
							// cout << "*** MoleculeAcceptor::accept: " << label << " immobilizing mol " << (*it).idString << " id " << mID.mnumber <<endl;
							m->immobilize((*it).immobilizeT);
						}

						if(m->getImmobilized() <= 1)
						{
							// register, who doomed the molecule
							m->setTakeoutPlace(label);
							
							// get rid of it
							// std::cerr << "*** MoleculeAcceptor::accept: cup = " << myWallsCup << endl;
							// std::cerr << "*** MoleculeAcceptor::accept: putting away with molecule " << (*it).idString << " id " << mID.mnumber << endl;
							
							if(myWallsCup) {
								// cout << "*** MoleculeAcceptor::accept: handing mol to cup" << endl;
								myWallsCup->doomMolecule(m);
								// cout << "*** MoleculeAcceptor::accept: handing mol to cup .. done" << endl;
							} else {
								cerr << "*** moleculeAcceptor::accept: NO CUP DEFINED -> molecule should be deleted directly ***" << endl;
							}
							// increase concentration of molecules in the reservoir
							// negative vol = count number of mols that shall reappear in the next insertion phase
							if((*it).invVol >= 0.0) {
								(*((*it).concentration)) += (*it).invVol;
							} else {
								// cout << "MoleculeAcceptor::accept: adding to number reservoir" << endl;
								(*((*it).concentration)) += 1.0;
								// cout << "MoleculeAcceptor::accept: dichte = " <<(*((*it).concentration)) << "  ("<<((*it).concentration) <<")"<<endl;
							}
							// check whether the full pos of the molecule should be saved
							if((*it).accPos) {
								(*(*it).accPos).push_back((m->getFullPosition()));
								fullPosition fp = m->getFullPosition();
								// cout << "## MoleculeAcceptor::accept: "<<label<<" accepting mol "<<(*it).idString<<", saving full position "<<fp.pos[0]<<" "<<fp.pos[1]<<" "<<fp.pos[2] << " (" <<fp.pos<<", "<< (*(*it).accPos).size() << " entries)" <<endl;
								// 							cout << "## MoleculeAcceptor::accept: "<<label<<" accepting mol "<<(*it).idString<<", saving full position "<<fp.pos[0]<<" "<<fp.pos[1]<<" "<<fp.pos[2]<<endl;
							}
							// std::cerr << "MoleculeAcceptor::accept: density = " << (*((*it).concentration)) << std::endl;
						
							return(true);
						}
					}
					// else
					// {
					// 	cout << "*** MoleculeAcceptor::accept: " << label << " is occupied: theInternal = " << *((*it).theInternal) << endl;
					// }
				}
			}
		} 
		// else do nothing, will be handed over to REFLECTIVE_WALL
		// std::cerr << "MoleculeAcceptor::accept: mol not inside mask" << std::endl;
		
		return(false);
	}
}
