// This file is part of the brownmove simulation package
// 
// (C) 2003 - 2009 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove

#include "proteinparser.h"

namespace brown {
	// access to global variable, defined in main.cpp
	extern double BETA;
	
	
	int ProteinParser::myGetLine(FILE *fp_in, char *newLine, const int maxLen)
	{
		int c;
		int count = 0;
		bool leadingWhite = true;
		bool ignoreLine = true;
		
		while(ignoreLine) {
			leadingWhite = true;
			count = 0;
			
			c = fgetc(fp_in);
			
			while(c != '\n')
			{
				if(c == EOF) {
					return(c);
				}
				
				if((c != ' ' && c != '\t')  || (leadingWhite == false))
				{
					leadingWhite = false;
					
					newLine[count] = c;
					count++;
					
					if(count >= maxLen) {
						newLine[count-1] = '\0';
						return(0);
					}
				}
				
				c = fgetc(fp_in);
			}
			newLine[count] = '\0';
			
			if(count > 0 && newLine[0] != '#') {
				ignoreLine = false;
			}
			
		}
		
		return(0);
	}
	
	
	int ProteinParser::firstWords(const char *newLine, char *tag, char *tag2)
	{
		int lang = strlen(newLine);
		int i = 0;

		if(sscanf(newLine, "%s %s", tag, tag2) == 2) {
			return(2);
		}
		else if(sscanf(newLine, "%s", tag) == 1) {
			tag2[0] = '\0';		// if only one word, return empty string as second arg
			return(1);
		} 
		else {
			return(0);
		}
	}
	
	
	 
	
	bool ProteinParser::parse(InteractingBody* ib, const char *filename, const int entry, const long int vdWMaxCol, list<externalForce> &felder)
	{
		char newLine[2048];
		char tag[2048], tag2[2048];
		int toBeSkipped = entry;
		bool opened = false;
		enum parserlevels level = OuterL;
		EstatShape *es;
		vdWShape *vs;
		HBondShape *hbs;
		BondShape *bs;
		GeomShape *gs;
		ExternalShape *xs;
		HiShape *his = NULL;	// set to NULL to test for second definition in the setup
		Gestalt *gest;
		Molecule *mol;
		Wall *wall;
		MoleculeAcceptor *mAcc = NULL;
		MoleculeInjector *mIn = NULL;
		std::map<std::string, long int> bondMap;
		bool bondsDone = false;
		enum bodyTypes IBType;
		
		// check if file can be opened
		FILE *fp_in;
		if(!(fp_in = fopen(filename, "r"))) {
			std::cerr << "ProteinParser::parse: error opening file >>" << filename << "<<" << std::endl;
			exit(-1);
		}
		
		std::cerr << "\nProteinParser::parse: Looking for item " << entry+1 << " in file " << filename << std::endl;
		
		bool stopper = true;
		while(stopper) 
		{
			if(myGetLine(fp_in, newLine, 2048) != EOF)
			{
				char proteinLabel[1024];
				fullPosition proteinPos; 
				int numOfTags;
				
				if((numOfTags = firstWords(newLine, tag, tag2)) == 0)
				{
					std::cerr << "ProteinParser::parse: error in line >>" << newLine << "<<" << std::endl;
					exit(-1);
				}
				
				if((strcmp(tag, "Protein") == 0) && (numOfTags == 2)) {
					opened = true;
					IBType = ProteinT;
					level = ProteinL;
					std::cerr << "ProteinParser::parse: found Protein " << tag2 << std::endl;
				}
				
				if((strcmp(tag, "Wall") == 0) && (numOfTags == 2)) {
					opened = true;
					IBType = WallT;
					level = ProteinL;
					std::cerr << "ProteinParser::parse: found Wall " << tag2 << std::endl;
				}

				// End of Protein => finalize
				if((strcmp(tag, "End") == 0 ) && (level == ProteinL)) 
				{
					// only finish object, if it has been set up (not skipped)
					if(toBeSkipped == 0)
					{
						std::vector<Gestalt>::iterator it;
						std::vector<Gestalt> *gstV;
						
						if(IBType == ProteinT) 
						{
							std::cerr << "ProteinParser::parse: ### found End of Protein " << proteinLabel << " ###" << std::endl;
							
							// Protein finished: teleport to new position
							ib->teleport(proteinPos);
							gstV = ib->getGestalten();
						}
						else if(IBType == WallT)
						{
							std::cerr << "ProteinParser::parse: ### found End of Wall " << proteinLabel << " ###" << std::endl;
							
							// place wall at final position
							ib->teleport(proteinPos);
							gstV = ib->getGestalten();
						}
						
						// some diagnostic output for the user
						{
							for(it = gstV->begin(); it != gstV->end(); it++) {
								
								std::cerr << "ProteinParser::parse: # Checking Gestalt: " << it->describe() << std::endl;
								if(it->getGeomShape()) {
									std::cerr << "ProteinParser::parse: #   GeomShape: " << it->getGeomShape()->describe() << std::endl;
								}
								if(it->getEstatShape()) {
									std::cerr << "ProteinParser::parse: #   EstatShape: " << it->getEstatShape()->describe() << std::endl;
								}
								if(it->getvdWShape()) {
									std::cerr << "ProteinParser::parse: #   vdWShape: " << it->getvdWShape()->describe() << std::endl;
								}
								if(it->getExternalShape()) {
									std::cerr << "ProteinParser::parse: #   ExternalShape: " << it->getExternalShape()->describe() << std::endl;
								}
								if(it->getHBondShape()) {
									std::cerr << "ProteinParser::parse: #   HBondShape: " << it->getHBondShape()->describe() << std::endl;
								}
								if(it->getBondShape()) {
									std::cerr << "ProteinParser::parse: #   BondShape: " << it->getBondShape()->describe() << std::endl;
								}
								if(it->getHiShape()) {
									std::cerr << "ProteinParser::parse: #   HiShape: " << it->getHiShape()->describe() << std::endl;
								}
							}
							
							if(IBType == ProteinT) {
								std::cerr << "ProteinParser::parse: ### finished Protein " << proteinLabel << " ###" << std::endl;
							}
							else if(IBType == WallT) {
								std::cerr << "ProteinParser::parse: ### finished Wall " << proteinLabel << " ###" << std::endl;
							}
						}
					}

					// one more protein/wall skipped
					if(toBeSkipped > 0) {
						toBeSkipped--;
					}
					else 	// we read the protein that we should, prepare for return
					{
						stopper = false;
					}
					
					opened = false;
					level = OuterL;
				}
				
				// #####################################################
				// Now we can read all the components to be assembled
				if(opened == true && toBeSkipped == 0)
				{
					char GestLabel[1024], tmpLabel[1024];
					fullPosition GestPos, tempPos; 
					position nullPos;
					double D0_t, D0_r, mass, radius;
					double D0_cm;

					doublevec_set(nullPos, 0.0);
					
					// here starts the protein
					if(strcmp(tag, "Protein") == 0)
					{
						strncpy(proteinLabel, tag2, 1024);
						proteinLabel[1023] = '\0';
						level = ProteinL;
						mol = (Molecule*) ib;		// specify that the IB is a molecule
						mol->setName(proteinLabel);
					} 
					else if(strcmp(tag, "Wall") == 0)
					{
						strncpy(proteinLabel, tag2, 1024);
						proteinLabel[1023] = '\0';
						level = ProteinL;
						wall = (Wall*) ib;			// the passed IB is a wall
						wall->setName(proteinLabel);
					} 
					else if((strcmp(tag, "Gestalt") == 0) && (numOfTags == 2))
					{
						strncpy(GestLabel, tag2, 1024);
						GestLabel[1023] = '\0';
						level = GestaltL;
						gest = new Gestalt(std::string(GestLabel));
						
						std::cerr << "ProteinParser::parse: Parsing Gestalt " << tag2 << std::endl;
					}
					else if((strcmp(tag, "EstatShape") == 0) && (numOfTags == 2))
					{
						strncpy(tmpLabel, tag2, 1024);
						tmpLabel[1023] = '\0';
						level = EstatShapeL;
						es = new EstatShape();
						
						std::cerr << "ProteinParser::parse: Parsing EstatShape " << tag2 << std::endl;
					}
					else if((strcmp(tag, "VdwShape") == 0) && (numOfTags == 2))
					{
						strncpy(tmpLabel, tag2, 1024);
						tmpLabel[1023] = '\0';
						level = vdWShapeL;
						vs = new vdWShape();
						
						std::cerr << "ProteinParser::parse: Parsing vdWShape " << tag2 << std::endl;
					}
					else if((strcmp(tag, "ExternalShape") == 0) && (numOfTags == 2))
					{
						strncpy(tmpLabel, tag2, 1024);
						tmpLabel[1023] = '\0';
						level = ExtShapeL;
						xs = new ExternalShape();
						
						std::cerr << "ProteinParser::parse: Parsing ExternalShape " << tag2 << std::endl;
					}
					else if((strcmp(tag, "HBondShape") == 0) && (numOfTags == 2))
					{
						strncpy(tmpLabel, tag2, 1024);
						tmpLabel[1023] = '\0';
						level = HBondShapeL;
						hbs = new HBondShape();
						
						std::cerr << "ProteinParser::parse: Parsing HBondShape " << tag2 << std::endl;
					}
					else if((strcmp(tag, "BondShape") == 0) && (numOfTags == 2))
					{
						strncpy(tmpLabel, tag2, 1024);
						tmpLabel[1023] = '\0';
						level = BondShapeL;
						bs = new BondShape();
						
						std::cerr << "ProteinParser::parse: Parsing BondShape " << tag2 << std::endl;
					}
					else if((strcmp(tag, "HiShape") == 0) && (numOfTags == 2))
					{
						strncpy(tmpLabel, tag2, 1024);
						tmpLabel[1023] = '\0';
						level = HiShapeL;
						
						std::cerr << "ProteinParser::parse: Parsing HiShape " << tag2 << std::endl;
					}
					else if(strcmp(tag, "Bonds") == 0)
					{
						if(bondsDone) {
							std::cerr << "ProteinParser::parse: too many bond definitions" << std::endl;
							exit(-1);
						}
						strncpy(tmpLabel, tag2, 1024);
						level = BondsL;
						std::cerr << "ProteinParser::parse: Parsing Bonds" << std::endl;
					}
					else if((strcmp(tag, "Injector") == 0) && (numOfTags == 2) && (IBType == WallT))		// Injector only for a wall
					{
						strncpy(tmpLabel, tag2, 1024);
						tmpLabel[1023] = '\0';
						level = InjectorL;
						if(mIn) {
							std::cerr << "ProteinParser::parse: second attempt to define an Injector..." << std::endl;
							exit(-1);
						}
						mIn = new MoleculeInjector(tag2);
						
						std::cerr << "ProteinParser::parse: Parsing Injector " << tag2 << std::endl;
					}
					else if((strcmp(tag, "Acceptor") == 0) && (numOfTags == 2) && (IBType == WallT))		// Acceptor only for a wall
					{
						strncpy(tmpLabel, tag2, 1024);
						tmpLabel[1023] = '\0';
						level = AcceptorL;
						if(mAcc) {
							std::cerr << "ProteinParser::parse: second attempt to define an Acceptor..." << std::endl;
							exit(-1);
						}
						mAcc = new MoleculeAcceptor(tag2);
						
						std::cerr << "ProteinParser::parse: Parsing Acceptor " << tag2 << std::endl;
					}
					// #### handle the various END tags depending on the current level #####
					//! @todo Verify that there are no empty shapes (might lead to trouble in the force determination, but is at least inefficient (call resp. shape))
					else if(strcmp(tag, "End") == 0)
					{
						switch(level) {
							// *** Protein finished: should never end here!!!
							case ProteinL:
							std::cerr << "ProteinParser::parse: premature END of protein???" << std::endl;
							exit(-1);
			 
							// *** Gestalt finished: teleport Gestalt and hook into protein
							case GestaltL:
							std::cerr << "ProteinParser::parse: found END(Gestalt)" << std::endl;
							gest->setMyPosition(GestPos);
							std::cerr << "ProteinParser::parse: set gestalt's position" << std::endl;
							gest->recalculate(GestPos);
							std::cerr << "ProteinParser::parse: recalculated Gestalt" << std::endl;
							ib->addGestalt(gest);
							std::cerr << "ProteinParser::parse: added Gestalt to molecule" << std::endl;
							
							delete his;		// reset HiShape pointer for next Gestalt
							his = NULL;
							
							level = ProteinL;	// return to out level
							std::cerr << "ProteinParser::parse: added Gestalt to molecule" << std::endl;
							break;
							
							// *** Bonds definition finished, reset level
							case BondsL:
							bondsDone = true;
							level = ProteinL;
							break;
							
							// *** Estatshape done: attach Estatshape to Gestalt
							case EstatShapeL:
							gest->setEstatShape(es);
							level = GestaltL;
							break;
							
							// *** vdWShape done: attach VdwShape to Gestalt
							case vdWShapeL:
							gest->setvdWShape(vs);
							level = GestaltL;
							break;
							
							// *** ExternalShape done: attach ExternalShape to Gestalt
							case ExtShapeL:
							gest->setExternalShape(xs);
							level = GestaltL;
							break;
							
							// *** HBondShape done
							case HBondShapeL:
							gest->setHBondShape(hbs);
							level = GestaltL;
							break;
							
							// *** BondShape done
							case BondShapeL:
							bs->setParameterVector(ib->getBondVector());
							gest->setBondShape(bs);
							level = GestaltL;
							break;
							
							// *** HiShape done: attach to Gestalt
							case HiShapeL:
							gest->setHiShape(his);
							level = GestaltL;
							break;
							
							// *** Injector done: attach to wall
							case InjectorL:
							wall->setInjector(mIn);
							level = ProteinL;
							break;
							
							// *** Acceptor done: attach to wall
							case AcceptorL:
							wall->setAcceptor(mAcc);
							level = ProteinL;
							break;
							
							// *** anything else: wrong ***
							default:
							std::cerr << "ProteinParser::parse: something wrong with >>" << newLine << "<<" << std::endl;
							exit(-1);
						}
					}
					// various uses of the position tag
					// BEWARE!! Rotation angles are given in degrees in the setup files!
					else if(strcmp(tag, "position") == 0)
					{
						position p;
						rotation r;
						
						if(sscanf(newLine, "position %lf %lf %lf %lf %lf %lf %lf", 
								  &p[0], &p[1], &p[2], &r.phi, &r.n[0], &r.n[1], &r.n[2]) != 7)
						{
							std::cerr << "ProteinParser::parse: cannot read position in line >>" << newLine << "<<" << std::endl;
							exit(-1);
						}
						r.phi *= (M_PI/180.0);
						std::cerr << "ProteinParser::parse: position = "<<p[0]<<", "<<p[1]<<", "<<p[2]<<" - "<<r.phi<<"rad, "<<r.n[0]<<", "<<r.n[1]<<", "<<r.n[2]<<std::endl;
						
						// save fullPos for InteractingBody, Gestalt, or for immediate use
						if(level == ProteinL) {
							doublevec_set(proteinPos.pos, p);
							rotation_set(proteinPos.rot, r);
						} else if(level == GestaltL) {
							doublevec_set(GestPos.pos, p);
							rotation_set(GestPos.rot, r);
						} else {
							doublevec_set(tempPos.pos, p);
							rotation_set(tempPos.rot, r);
						}
					}
					else if(strcmp(tag, "D0") == 0)
					{
						// different meanings at Gestalt and protein level
						if(level == GestaltL)
						{
							if(sscanf(newLine, "D0 %lf %lf %lf %lf", &D0_t, &D0_r, &mass, &radius) != 4) {
								std::cerr << "ProteinParser::parse: Error reading >>" << newLine << "<<" << std::endl;
								exit(-1);
							}
							gs = new GeomShape();
							
							gs->setDTrans(D0_t, timeStep, BETA, mass, true);
							if(D0_r > 0.0) {
								gs->setDoesRotate(true);
								gs->setDRot(D0_r, timeStep, BETA, mass, radius, true);
							} else {
								gs->setDoesRotate(false);
							}
							
							gest->setGeomShape(gs);
						}
						// D0 of the complete protein is required by a moleculeinjector
						else if(level == ProteinL)
						{
							if(sscanf(newLine, "D0 %lf", &D0_cm) != 1) {
								std::cerr << "ProteinParser::parse: Error reading >>" << newLine << "<<" << std::endl;
								exit(-1);
							}
							mol->setA0Trans(sqrt(4.0 * D0_cm * timeStep));
						}
					}
					// ### anything else is data with a level-dependent interpretation ###
					else
					{
						switch(level) {
							case BondsL:
							{
								double R_0, k;
								long int idx;
								
								if(sscanf(newLine, "%s %lf %lf", tmpLabel, &R_0, &k) != 3) {
									std::cerr << "ProteinParser::parse: error in bond definition line >>" << newLine << "<<" << std::endl;
									exit(-1);
								}
								idx = ib->addBondParameter(R_0, k, std::string(tmpLabel));
								bondMap[std::string(tmpLabel)] = idx;
								std::cerr << "ProteinParser::parse: added parameters for bond " << tmpLabel << " as #" << idx << "  R_0 = " << R_0 << ",  k = " << k << std::endl;
							}
							break;
							
							case EstatShapeL:
							{
								chargePos cpos;
								if(sscanf(newLine, "%lf %lf %lf %lf %lf", 
										  &cpos.pos[0], &cpos.pos[1], &cpos.pos[2], &cpos.ch, &cpos.r) != 5) {
									std::cerr << "ProteinParser::parse: error in chargePos line >>"  << newLine << "<<" << std::endl;
									exit(-1);
								}
								es->addCharge(cpos, nullPos);
							}
							break;
							
							case ExtShapeL:
							{
								externalPos epos;
								bool foundField = false;
								if(sscanf(newLine, "%lf %lf %lf %s %lf", 
										  &epos.pos[0], &epos.pos[1], &epos.pos[2], tmpLabel, &epos.alpha) != 5) {
									std::cerr << "ProteinParser::parse: error in externalPos line >>"  << newLine << "<<" << std::endl;
									exit(-1);
								}
								
								for(list<externalForce>::iterator it=felder.begin(); it != felder.end(); ++it) {
									if((*it).label == string(tmpLabel)) {
										foundField = true;
										epos.setField((*it));
										break;
									}
								}
								if(foundField) {
									xs->addHook(epos, nullPos);
								}
								else {
									// work-around for extractor: if there is a single external field of type NONE, do not issue a warning
									if(!(felder.size() == 1 && felder.end()->type == NONE)) {
										std::cerr << "ProteinParser::parse: could not identify external field >>" << tmpLabel << "<<" << std::endl;
										exit(-1);
									}
									
								}
							}
							break;
							
							case vdWShapeL:
							{
								// If it's a wall: check for wall or box
								// If all cases fail, the attempt to read vdwPos will spit out the error :-)
								vdWBlock vbl;
								if(IBType == WallT) 
								{
									double offset;
									long int clr;
									if (sscanf(newLine, "wall %lf %ld", &offset, &clr) == 2) {
										if((clr >= 0) && (clr < vdWMaxCol)) {
											std::cerr << "Proteinparser::parse: found vdw wall with offset " << offset << " and color " << clr << std::endl;
											vs->setWall(offset, clr);
										} else {
											std::cerr << "Proteinparser::parse: vdW color " << clr << " is outside bounds [0 .. " << vdWMaxCol-1 << "]" << std::endl;
											exit(-1);
										}
										break;
									}
									else if (sscanf(newLine, "block %lf %lf %lf %lf %lf %lf %lf %ld", 
													&vbl.pos[0], &vbl.pos[1], &vbl.pos[2], &vbl.lx, &vbl.ly, &vbl.lz, &vbl.r, &vbl.color) == 8)
									{
										if((vbl.color >= 0) && (vbl.color < vdWMaxCol)) {
											std::cerr << "Proteinparser::parse: found vdw block at "<<vbl.pos[0]<<" "<<vbl.pos[1]<<" "<<vbl.pos[2]<<"  with dimensions " << vbl.lx<<" "<<vbl.ly<<" "<<vbl.lz<<", radius " << vbl.r << ", and color " << vbl.color << std::endl;
											vs->addBlock(vbl);
										} else {
											std::cerr << "Proteinparser::parse: vdW color " << vbl.color << " is outside bounds [0 .. " << vdWMaxCol-1 << "]" << std::endl;
											exit(-1);
										}
										break;
									}
								}
								
								vdWPos vpos;
								if(sscanf(newLine, "%lf %lf %lf %lf %ld", &vpos.pos[0], &vpos.pos[1], &vpos.pos[2], &vpos.r, &vpos.color) == 5) 
								{
									if((vpos.color >= 0) && (vpos.color < vdWMaxCol)) {
										vs->addSphere(vpos, nullPos);		  
									} else {
										std::cerr << "Proteinparser::parse: vdW color " << vpos.color << " is outside bounds [0 .. " << vdWMaxCol-1 << "] in line " << newLine << std::endl;
										exit(-1);
									}
								}
								else {
									std::cerr << "ProteinParser::parse: error in vdWPos line >>"  << newLine << "<<" << std::endl;
									exit(-1);
								}
							}
							break;
							
							case HBondShapeL:
							{
								hbondPos hbpos;
								double lang;
								
								// try for acceptor: position of acceptor atom
								if(sscanf(newLine, "acceptor %lf %lf %lf %lf", 
										  &hbpos.pos[0], &hbpos.pos[1], &hbpos.pos[2], &hbpos.giveOrTake) == 4) {
									if(hbpos.giveOrTake <= 0.0) {
										std::cerr << "ProteinParser::parse: scaling factor must be positive in >>" << newLine << "<<" << std::endl;
										exit(-1);
									}
									hbpos.giveOrTake *= -1.0;
									// set direction to zero
									doublevec_set(hbpos.bond, 0.0);
									// hook up
									hbs->addHook(hbpos, nullPos);
								} 
								// or donor: position of donor and hydrogen => position of hydrogen and vector donor->H
								else if(sscanf(newLine, "donor %lf %lf %lf %lf %lf %lf %lf", 
											   &hbpos.bond[0], &hbpos.bond[1], &hbpos.bond[2], &hbpos.pos[0], &hbpos.pos[1], &hbpos.pos[2], &hbpos.giveOrTake) == 7) {
									if(hbpos.giveOrTake <= 0.0) {
										std::cerr << "ProteinParser::parse: scaling factor must be positive in >>" << newLine << "<<" << std::endl;
										exit(-1);
									}
									// normalize and change direction
									doublevec_sub(hbpos.bond, hbpos.pos);
									lang = doublevec_length(hbpos.bond);
									doublevec_mult(hbpos.bond, (-1.0/lang));
									// hook up
									hbs->addHook(hbpos, nullPos); 
								}
								// or something is wrong
								else
								{
									std::cerr << "ProteinParser::parse: error in hbond line >>"  << newLine << "<<" << std::endl;
									exit(-1);
								}
							}
							break;
							
							case BondShapeL:
							{
								bondPos bp;
								if(sscanf(newLine, "%lf %lf %lf %s", 
										  &bp.pos[0], &bp.pos[1], &bp.pos[2], tmpLabel) != 4) {
									std::cerr << "ProteinParser::parse: error in bondPos line >>"  << newLine << "<<" << std::endl;
									exit(-1);
								}
								if(bondMap.count(std::string(tmpLabel)) == 1) {
									bp.idx = bondMap[std::string(tmpLabel)];
									bs->addHook(bp, nullPos);
									std::cerr << "ProteinParser::parse: added bond hook to bond " << tmpLabel << std::endl;
								} else {
									std::cerr << "ProteinParser::parse: Bond >>" << tmpLabel << "<< is not defined" << std::endl;
									exit(-1);
								}
							}
							break;
							
							case HiShapeL:
							{
								double radius;
								if(sscanf(newLine, "%lf", &radius) != 1) {
									std::cerr << "ProteinParser::parse: error in HiShape line >>" << newLine << "<<" << std::endl;
									exit(-1);
								}
								if(his) {
									std::cerr << "ProteinParser::parse: only one HiShape radius definition allowed!" << std::endl;
									exit(-1);
								}
								
								his = new HiShape(radius);
							}
							break;
							
							case InjectorL:
							{
								char proteinName[1024], reservoirName[1024];
								double maskY, maskZ, maskdY, maskdZ;
								
								if(sscanf(newLine, "Mask %lf %lf %lf %lf", &maskY, &maskZ, &maskdY, &maskdZ) == 4) {
									mIn->setRectangleMask(maskY, maskZ, maskdY, maskdZ);
								} 
								else if(sscanf(newLine, "Molecule %s %s", proteinName, &reservoirName) == 2)  {
									mIn->addMolecule(std::string(proteinName), std::string(reservoirName));
								}
								else
								{
									std::cerr << "ProteinParser::parse: error in Injector line >>" << newLine << "<<" << std::endl;
									exit(-1);
								}
							}
							break;
							
							case AcceptorL:
							{
								char proteinName[1024], reservoirName[1024];
								double maskY, maskZ, maskdY, maskdZ;
								
								if(sscanf(newLine, "Mask %lf %lf %lf %lf", &maskY, &maskZ, &maskdY, &maskdZ) == 4) {
									mAcc->setRectangleMask(maskY, maskZ, maskdY, maskdZ);
								} 
								else if(sscanf(newLine, "Molecule %s %s", proteinName, &reservoirName) == 2) {
									mAcc->addMolecule(std::string(proteinName), std::string(reservoirName));
								}
								else
								{
									std::cerr << "ProteinParser::parse: error in Acceptor line >>" << newLine << "<<" << std::endl;
									exit(-1);
								}
							}
							break;
							
							default:
							std::cerr << "ProteinParser::parse: we shouldn't be here: >>" << newLine << "<<" << std::endl;
							exit(-1);
						}
					}
				}
				// #####################################################
			}
			else
			{
				std::cerr << "ProteinParser::parse: reached EOF" << std::endl;
				stopper = false;
				fclose(fp_in);
				return(false);
			}
		}
		fclose(fp_in);

		// at the end of the quest:
		return(true);
	}
}
