// This file is part of the brownmove simulation package
// 
// (C) 2003 - 2009 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove

#include <iostream>
#include <fstream>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "constants.h"
#include "types.h"
#include "vectormath.h"
#include "random.h"

#include "cup.h"
#include "wall.h"
#include "molecule.h"


#include "proteinparser.h"


using namespace brown;

namespace brown {
	// some global variables for global stuff
	// Q: Should we add the timestep here? => No, as long as it does not change during the simulation
	// 
	extern double KAPPA;			// for physiological salt concentration of 0.09M
	extern double EPS_H2O;
	extern double BETA;				// default temp = 20 degrees
	
	extern double HBOND_DEPTH;		// average H-bond: 25 kJ/mol
	extern double HBOND_LENGTH;		// length of an average H bond (between donor-hydrogen and acceptor-atom)
	extern double HBOND_WIDTH;		// potential width in nm
	extern double HBOND_CUTOFF;		// ignore H bond if donor and acceptor are further apart
}


// Read next line, chop off leading whitespace and ignore comment lines
int myGetLine(FILE *fp_in, char *newLine, const int maxLen)
{
	int c;
	int count = 0;
	bool leadingWhite = true;
	bool ignoreLine = true;
	
	while(ignoreLine) {
		leadingWhite = true;
		count = 0;
		
		c = fgetc(fp_in);
		
		while(c != '\n') {
			if(c == EOF) {
				return(c); }
			
			if((c != ' ' && c != '\t')  || leadingWhite == false) {
				leadingWhite = false;
				newLine[count] = c;
				count++;
				if(count >= maxLen) {
					newLine[count-1] = '\0';
					return(0);
				}
			}
			c = fgetc(fp_in);
		}
		newLine[count] = '\0';
		if(count > 0 && newLine[0] != '#') {
			ignoreLine = false; }
	}
	return(0);
}


int firstWords(const char *newLine, char *tag, char *tag2)
{
	int lang = strlen(newLine);
	int i = 0;

	if(sscanf(newLine, "%s %s", tag, tag2) == 2) {
		return(2); }
	else if(sscanf(newLine, "%s", tag) == 1) {
		tag2[0] = '\0';		// return empty string as second arg
		return(1); } 
	else {
		return(0);
	}
}


// set up random etc, 
void initialize() {
	cerr << "### initialize: random generator" << endl;
	init_ran();
}


// Init the cup
void init_cup(Cup *theCup, char *setupFileName, char *commandline, double &timeStep, long int &dauer, long int &ausgabe, long int &Offset, bool &noDump, char *outputFile, list<string> &densities, list<string> &numbers, list<externalForce> &felder)
{
	int i;
	fullPosition fp;
	
	// +++++ create the cup +++++++
	cerr << "### init_cup: set up the cup" << endl;
	theCup->initialize();
	

	// +++++ Open setup file and parse +++++
	FILE *fp_setup;
	char newLine[2048];
	char tag[2048], tag2[2048];
	char helfer[2048], helfer2[2048];
	char sceneFileName[2048];
	char proteinFileName[2048];
	long int rein1, rein2, vdWMaxCol = 0;
	double dein1, dein2, dein3;
	vdWParameter vdWPara;
	double HICutOff = 1.0e20;
	bool sceneDefined = false;
	std::map<std::string, Molecule*> molShelf;
	std::list<std::string> definedReservoirs;
	std::list<std::string> definedParticles;
	
	// set to zero for test whether an outputfile was defined
	outputFile[0] = '\0';
	
	if(!(fp_setup = fopen(setupFileName, "r")))
	{
		std::cerr << "main::init_cup: cannot open setup file >>" << setupFileName << "<<" << std::endl;
		exit(-1);
	}
	std::cerr << "main::init_cup: Reading setup from setup file >>" << setupFileName << "<<" << std::endl;
	

	ProteinParser *p = NULL;
	std::map<std::string, Molecule*> molMap;

	// parse the setup file
	bool stopper = true;
	while(stopper)
	{
		if(myGetLine(fp_setup, newLine, 2048) != EOF)
		{
			int numOfTags = firstWords(newLine, tag, tag2);
								   
			// timestep
			if(sscanf(newLine, "timestep %lf", &dein1) == 1) {
				if(dein1 > 0) {
					timeStep = dein1;
					std::cerr << "main::init_cup: timestep = " << timeStep << " ps" << std::endl;
					p = new ProteinParser(timeStep);
				} else {
					std::cerr << "main_init_cup: timestep " << dein1 << " <= 0" << std::endl;
					exit(-1);
				}
			} 
			// simulation length
			else if(sscanf(newLine, "runsteps %ld", &rein1) == 1) {
				if(rein1 > 0) {
					dauer = rein1;
					std::cerr << "main::init_cup: duration = " << dauer << " timesteps = " << timeStep * dauer * 1e-6 << " micro secs" << std::endl;
				} else {
					std::cerr << "main_init_cup: simulation length " << rein1 << " <= 0" << std::endl;
					exit(-1);
				}
			}
			// output interval
			else if(sscanf(newLine, "outsteps %ld", &rein1) == 1) {
				if(rein1 > 0) {
					ausgabe = rein1;
					std::cerr << "main::init_cup: output every " << ausgabe << " timesteps = " << timeStep * ausgabe << " ps" << std::endl;
				} else {
					std::cerr << "main_init_cup: output interval " << rein1 << " <= 0" << std::endl;
					exit(-1);
				}
			}
			// optional offset of timesteps (continued from a previous run), might even be <= 0
			else if(sscanf(newLine, "offsetsteps %ld", &rein1) == 1) {
				Offset = rein1;
				std::cerr << "main::init_cup: run sim with an offset of " << Offset << " timesteps = " << timeStep * Offset << " ps" << std::endl;
			}
			// optional random seed
			else if(sscanf(newLine, "randomseed %ld", &rein1) == 1) {
				unsigned long int seed = (unsigned long int)(rein1);
				random_seed(seed);
				std::cerr << "main::init_cup: ##### seeded random generator with " << seed << " #####" << std::endl;
			}
			// optional: do not dump the proteins in the end
			else if(strcmp(tag, "noProteinDump") == 0) {
				noDump = true;
				std::cerr << "main::init_cup: no final dump of the proteins" << std::endl;
			}
			// vdW interaction parameters
			else if(sscanf(newLine, "vdwarraysize %ld", &vdWMaxCol) == 1) {
				if(vdWMaxCol > 0) {
					theCup->initializeVDWArray(vdWMaxCol);
					std::cerr << "main::init_cup: preparing to use " << vdWMaxCol << " vdW color" << ((vdWMaxCol>1)?"s":"") << std::endl;
				} else {
					std::cerr << "main::init_cup: >>" << vdWMaxCol << " should be >= 1" << std::endl;
					exit(-1);
				}
			}
			// vdW parameters
			else if(sscanf(newLine, "vdwparms %ld %ld %lf %lf %lf %lf %lf %lf %lf", 
						   &rein1, &rein2, &vdWPara.C_12, &vdWPara.C_6, &vdWPara.C_3, &vdWPara.R_0, &vdWPara.DR, &vdWPara.Rmin, &vdWPara.Rmax) == 9) {
				if((rein1 < vdWMaxCol) && (rein2 < vdWMaxCol)) {
					theCup->setVDWParam(rein1, rein2, vdWPara);
					std::cerr << "main::init_cup: setting vdW params for " << rein1 << ", " << rein2 << std::endl;
				} else {
					std::cerr << "main::init_cup: vdW colors to large: " << rein1 <<", " << rein2 << " > " << vdWMaxCol-1 << std::endl;
					exit(-1);
				}
			}
			// optional HI cutoff			   
			else if(sscanf(newLine, "hicutoff %lf", &dein1) == 1) {
				HICutOff = dein1;
				std::cerr << "main::init_cup: using HI cutoff = " << HICutOff << std::endl;
			}
			// electrostatic shielding: KAPPA = 1/DebyeLength
			else if(sscanf(newLine, "kappa %lf", &dein1) == 1) {
				if(dein1 >= 0.0) {
					brown::KAPPA = dein1;
					std::cerr << "main::init_cup: using Coulomb shielding parameter kappa = " << brown::KAPPA << " nm^-1" << std::endl;
				} else {
					std::cerr << "main::init_cup: Coulomb shielding parameter kappa = " << dein1 << " < 0.0" << std::endl;
					exit(-1);
				}
			}
			// relative dielectric constant of water
			else if(sscanf(newLine, "eps_H2O %lf", &dein1) == 1) {
				if(dein1 > 0.0) {
					EPS_H2O = dein1;
					std::cerr << "main::init_cup: using relative dielectric constant of water eps_H2O = " << EPS_H2O << std::endl;
				} else {
					std::cerr << "main::init_cup: relative dielectric constant of water eps_H2O = " << dein1 << " <= 0.0" << std::endl;
					exit(-1);
				}
			}
			// temperature of the solvent
			else if(sscanf(newLine, "temperature %lf", &dein1) == 1) {
				if(dein1 > 0.0) {
					BETA = 1.0/(K_B * dein1);
					std::cerr << "main::init_cup: using solvent temperature = " << dein1 << " K" << std::endl;
				} else {
					std::cerr << "main::init_cup: solvent temperature = " << dein1 << " <= 0.0" << std::endl;
					exit(-1);
				}
			}
			// optional HBond parameters: potential depth
			else if(sscanf(newLine, "hbond_depth %lf", &dein1) == 1) {
				if(dein1 > 0.0) {
					HBOND_DEPTH = dein1;
					std::cerr << "main::init_cup: using hbond potential depth = " << dein1 << " kJ/mol" << std::endl;
				} else {
					std::cerr << "main::init_cup: hbond potential depth = " << dein1 << " <= 0.0" << std::endl;
					exit(-1);
				}
			}
			// optional HBond parameters: length between H and acceptor
			else if(sscanf(newLine, "hbond_length %lf", &dein1) == 1) {
				if(dein1 > 0.0) {
					HBOND_LENGTH = dein1;
					std::cerr << "main::init_cup: using hbond length (h to acceptor) = " << dein1 << " nm" << std::endl;
				} else {
					std::cerr << "main::init_cup: hbond length = " << dein1 << " <= 0.0" << std::endl;
					exit(-1);
				}
			}
			// optional HBond parameters: potential width
			else if(sscanf(newLine, "hbond_width %lf", &dein1) == 1) {
				if(dein1 > 0.0) {
					HBOND_WIDTH = dein1;
					std::cerr << "main::init_cup: using hbond potential width = " << dein1 << " nm" << std::endl;
				} else {
					std::cerr << "main::init_cup: hbond potential width = " << dein1 << " <= 0.0" << std::endl;
					exit(-1);
				}
			}
			// optional HBond parameters: cutoff distance
			else if(sscanf(newLine, "hbond_cutoff %lf", &dein1) == 1) {
				if(dein1 > 0.0) {
					HBOND_CUTOFF = dein1;
					std::cerr << "main::init_cup: using hbond cutoff = " << dein1 << " nm" << std::endl;
				} else {
					std::cerr << "main::init_cup: hbond cutoff = " << dein1 << " <= 0.0" << std::endl;
					exit(-1);
				}
			}

			// analyzer script
			else if(strcmp(tag, "analysis") == 0) {
				// copy newLine into analysis script line from char 9 on
				int i = 9;
				int k = 0;
				bool leadingSpace = true;
				while(newLine[i] != '\0') 
				{
					if((leadingSpace == false) || (newLine[i] != ' ')) {
						leadingSpace = false;
						commandline[k] = newLine[i];
						k++;
					}
					i++;
				}
				commandline[k] = '\0';
			} 
			
			// definition of external forces
			else if(sscanf(newLine, "externalForce %s %s %lf %lf %lf", helfer, helfer2, &dein1, &dein2, &dein3) == 5) {
				externalForce extF;
				extF.label = helfer;
				
				// check type
				if(strcmp(helfer2, "constant") == 0) {		// constant force field
					extF.type = LINEAR;
					doublevec_set(extF.dir, dein1, dein2, dein3);
				}
				else if(strcmp(helfer2, "harmonic") == 0) {		// harmonic potential
					extF.type = HARMONIC;
					if(sscanf(newLine, "externalForce %s %s %lf %lf %lf %lf %lf %lf", 
							  helfer, helfer2, &extF.pos[0], &extF.pos[1], &extF.pos[2], &extF.dir[0], &extF.dir[1], &extF.dir[2]) != 8) {
						std::cerr << "main::init_cup: error reading >>" << newLine << "<< into harmonic external force" << std::endl;
						exit(-1);
					}
				}
				else if(strcmp(helfer2, "shear") == 0) {		// shear flow
					extF.type = SHEAR;
					if(sscanf(newLine, "externalForce %s %s %lf %lf %lf %lf %lf %lf %lf %lf %lf", 
							  helfer, helfer2, &extF.pos[0], &extF.pos[1], &extF.pos[2], &extF.dir[0], &extF.dir[1], &extF.dir[2], &extF.flow[0], &extF.flow[1], &extF.flow[2]) != 11) {
						std::cerr << "main::init_cup: error reading >>" << newLine << "<< into shear force" << std::endl;
						exit(-1);
					}
				}
				else
				{
					std::cerr << "main::init_cup: unknown type >>" << helfer2 << "<< for external force" << std::endl;
					exit(-1);
				}
				felder.push_back(extF);
				std::cerr << "main::init_cup: set up external force >>" << extF.label << "<< of type >>" << helfer2 << "<<" << std::endl;
			}
			
			
			// read protein definitions
			else if(sscanf(newLine, "protein %s %s %ld", helfer, helfer2, &rein1) == 3)
			{
				strncpy(proteinFileName, helfer2, 2047);
				
				// check if parser was already initialized
				if(!p) {
					std::cerr << "main::init_cup: timestep definition must come before protein definition!!!" << std::endl;
					exit(-1);
				}
				
				Molecule *m1 = new Molecule("temp", 0);

				if(!(p->parse(m1, proteinFileName, (rein1-1), vdWMaxCol, felder))) {
					std::cerr << "main::init_cup: error parsing molecule definition " << rein1 << " in file >>" << proteinFileName << std::endl;
					exit(-1);
				}
				std::cerr << "main::init_cup: molecule " << (m1->getid()).mclass << " done" << std::endl;
				
				// save for potential output counting
				definedParticles.push_back((m1->getid()).mclass);
				
				std::cerr << "main::init_cup: molecule: HICutOff = " << HICutOff << std::endl;
				m1->setHICutOff(HICutOff);
				
				molMap[std::string((m1->getid()).mclass)] = m1;
				molShelf[std::string(helfer)] = m1;
			}
			
			// definition of the simulation box = walls
			else if(sscanf(newLine, "box %s", helfer) == 1) 
			{
				if(sceneDefined) {
					std::cerr << "main::init_cup: another scenery definition in line >>" << newLine << "<<" << std::endl;
				}
				sceneDefined = true;

				strncpy(sceneFileName, helfer, 2047);
				
				// check if parser was already initialized
				if(!p) {
					std::cerr << "main::init_cup: timestep definition must come before box definition!!!" << std::endl;
					exit(-1);
				}
				std::cerr << "main::init_cup: reading walls from box definition file >>" << sceneFileName << "<<" << std::endl;
				
				// parse walls
				int iw = 0;
				while(true)
				{
					Wall *w = new Wall("temp", 0);
					
					if(p->parse(w, sceneFileName, iw, vdWMaxCol, felder)) 
					{
						if(w->getInjector()) {
							w->getInjector()->resolveMolNames(&molMap);
						}
						std::cerr << "main::init_cup: mol names resolved" << std::endl;
						theCup->addWall(*w);
						std::cerr << "main::init_cup: wall added to cup" << std::endl;
						
						iw++;
					}
					else {
						break;
					}
					
					delete w;	// free memory again
				}
				std::cerr << "main::init_cup: done with walls\n" << std::endl;
				
				theCup->listWalls();
				
				std::cerr << "main::init_cup: listing done" << std::endl;
			}
			
			// definition of a (quasi) periodic scenery
			else if(sscanf(newLine, "periodicBox %lf %lf %lf", &dein1, &dein2, &dein3) == 3)
			{
				if(sceneDefined) {
					std::cerr << "main::init_cup: duplicate scenery definition in line >>" << newLine << "<<" << std::endl;
					exit(-1);
				}
				sceneDefined = true;
				
				double Xper, Yper, Zper;
				fullPosition fp;
				
				Xper = dein1; Yper = dein2; Zper = dein3;
				
				std::cerr << "main::init_cup: setting up periodic box with dimensions " << Xper << ", "<<Yper<<", " << Zper << std::endl;
				
				theCup->setUnitCellSize(Xper, Yper, Zper);
				
				Wall *w;
				Wall::WALLTYPES wt = Wall::PERIODIC_WALL;
				
				if(Xper > 0.0) {
					w = new Wall("Xminus", 0);
					w->setWallType(wt);
					w->setPeriodicShift(Xper, 0.0, 0.0);
					doublevec_set(fp.pos, -0.5*Xper, 0.0, 0.0);
					rotation_set(fp.rot, 0.0, 1.0, 0.0, 0.0);
					w->teleport(fp);
					theCup->addWall(*w);
					delete w;
					
					w = new Wall("Xplus", 0);
					w->setWallType(wt);
					w->setPeriodicShift(-Xper, 0.0, 0.0);
					doublevec_set(fp.pos, 0.5*Xper, 0.0, 0.0);
					rotation_set(fp.rot, M_PI, 0.0, 0.0, 1.0);
					w->teleport(fp);
					theCup->addWall(*w);
					delete w;
				}
				
				if(Yper > 0.0) {
					w = new Wall("Yminus", 0);
					w->setWallType(wt);
					w->setPeriodicShift(0.0, Yper, 0.0);
					doublevec_set(fp.pos, 0.0, -0.5*Yper, 0.0);
					rotation_set(fp.rot, 0.5*M_PI, 0.0, 0.0, 1.0);
					w->teleport(fp);
					theCup->addWall(*w);
					delete w;
					
					w = new Wall("Yplus", 0);
					w->setWallType(wt);
					w->setPeriodicShift(0.0, -Yper, 0.0);
					doublevec_set(fp.pos, 0.0, 0.5*Yper, 0.0);
					rotation_set(fp.rot, -0.5*M_PI, 0.0, 0.0, 1.0);
					w->teleport(fp);
					theCup->addWall(*w);
					delete w;
				}
				
				if(Zper > 0.0) {
					w = new Wall("Zminus", 0);
					w->setWallType(wt);
					w->setPeriodicShift(0.0, 0.0, Zper);
					doublevec_set(fp.pos, 0.0, 0.0, -0.5*Zper);
					rotation_set(fp.rot, -0.5*M_PI, 0.0, 1.0, 0.0);
					w->teleport(fp);
					theCup->addWall(*w);
					delete w;
					
					w = new Wall("Zplus", 0);
					w->setWallType(wt);
					w->setPeriodicShift(0.0, 0.0, -Zper);
					doublevec_set(fp.pos, 0.0, 0.0, 0.5*Zper);
					rotation_set(fp.rot, 0.5*M_PI, 0.0, 1.0, 0.0);
					w->teleport(fp);
					theCup->addWall(*w);
					delete w;
				}
				
				std::cerr << "main::init_cup: setup done" << std::endl;
			}
			
			// Reservoirs
			else if(sscanf(newLine, "reservoir %s %lf %lf", helfer, &dein1, &dein2) == 3)
			{
				if(dein1 < 0.0) {
					std::cerr << "main::init_cup: reservoir " << helfer << " has initial density " << dein1 << " < 0.0" << std::endl;
					exit(-1);
				}
				std::cerr << "main::init_cup: found reservoir " << helfer << " with initial density " << dein1 << " and inv. vol " << dein2 << std::endl;
				theCup->addReservoir(helfer, dein1, dein2);
				definedReservoirs.push_back(helfer);
			}
			
			// initially placed proteins
			else if(sscanf(newLine, "startWith %s %lf %lf %lf %lf %lf %lf %lf", 
						   helfer, &fp.pos[0], &fp.pos[1], &fp.pos[2], &fp.rot.phi, &fp.rot.n[0], &fp.rot.n[1], &fp.rot.n[2]) == 8)
			{
				fp.rot.phi *= (M_PI / 180.0);	// convert to degrees
				
				if(molShelf.count(std::string(helfer)) == 1)
				{
					Molecule *mm = molShelf[std::string(helfer)];
					mm->teleport(fp);
					theCup->addMolecule(*mm);
					std::cerr << "main::init_cup: placing protein " << helfer << " at position " << fp.pos[0]<<" "<<fp.pos[1]<<" "<<fp.pos[2]<<" "<<fp.rot.phi<<" "<<fp.rot.n[0]<<" "<<fp.rot.n[1]<<" "<<fp.rot.n[2] << std::endl;
				}
				else 
				{
					std::cerr << "main::init_cup: molecule " << helfer << " was not set up" << std::endl;
					exit(-1);
				}
			}
			// output of the reservoir densities and particle numbers
			else if(sscanf(newLine, "outputfile %s", helfer) == 1)
			{
				strncpy(outputFile, helfer, 2047);
				std::cerr << "main::init_cup: registered output file >>" << outputFile << "<<" << std::endl;
			}
			
			else if(sscanf(newLine, "density %s", helfer) == 1)
			{
				if(strlen(outputFile) > 0) {	// filename is defined
					// test whether this reservoir was already defined
					list<string>::iterator it;
					bool isDefined = false;
					for(it = definedReservoirs.begin(); it != definedReservoirs.end(); ++it) {
						if((*it) == string(helfer)) {
							isDefined = true;
							break;
						}
					}
					if(isDefined) {
						std::cerr << "main::init_cup: registered reservoir >>" << helfer << "<< for output" << std::endl;
						densities.push_back(helfer);
					} else {
						std::cerr << "main::init_cup: error registering reservoir >>" << helfer << "<< for output (not defined?)" << std::endl;
						exit(-1);
					}
				}
				else {
					std::cerr << "main::init_cup: no output file defined, cannot register reservoir >>" << helfer << "<<" << std::endl;
					exit(-1);
				}
			}
			
			else if(sscanf(newLine, "number %s", helfer) == 1)
			{
				if(strlen(outputFile) > 0) {	// filename is defined
					// test whether this protein type was already defined
					list<string>::iterator it;
					bool isDefined = false;
					for(it = definedParticles.begin(); it != definedParticles.end(); ++it) {
						if((*it) == string(helfer)) {
							isDefined = true;
							break;
						}
					}
					if(isDefined) {
						std::cerr << "main::init_cup: registered protein type >>" << helfer << "<< for output" << std::endl;
						numbers.push_back(helfer);
					} else {
						std::cerr << "main::init_cup: error registering protein type >>" << helfer << "<< for output (not defined?)" << std::endl;
						exit(-1);
					}
				}
				else {
					std::cerr << "main::init_cup: no output file defined, cannot register protein type >>" << helfer << "<<" << std::endl;
					exit(-1);
				}
			}

			
			// anything that does fit not any of the above tests must be wrong
			else {
				std::cerr << "main::init_cup: cannot make sense of setup line >>" << newLine << "<<" << std::endl;
				exit(-1);
			}
		}
		// We're done, jipii!
		else
		{
			std::cerr << "main::init_cup: reached EOF" << std::endl;
			stopper = false;
			fclose(fp_setup);
		}
		
	}
	
	std::cerr << "main::init_cup: setup done" << std::endl;
	
	// last check: how many molecules
	std::cerr << "init_cup: number of mols in the cup = " << theCup->countMolecules() << endl;
	// theCup->listMolecules();
	
	std::cerr << "### init_cup: finished ################" << endl;
	std::cerr << "#######################################" << endl;
}


// propagate the molecules
void main_loop(Cup *theCup, FILE *fp_a, double Time_Step,long Length, long Out_var, long Offset, bool noDump, char *outputFile, list<string> &densities, list<string> &numbers)
{
	timestamp it, i_out, i_report;
	long int outCounter;
	FILE *fp_d;
	i_out = 1;
	i_report = 1;
	outCounter = 0;
	
	// if defined, open file for output of reservoir densities
	if(strlen(outputFile) > 0) {
		if(!(fp_d = fopen(outputFile, "w"))) {		// can't open file
			std::cerr << "main_loop: Error opening output file >>" << outputFile << "<<" << std::endl;
			exit(-1);
		}
		else
		{
			// print Header: "#t[\treservoirNames][\tparticleTypes]
			fprintf(fp_d, "#time");
			
			std::list<std::string>::iterator it;
			for(it = densities.begin(); it != densities.end(); ++it) {
				fprintf(fp_d, "\t%s", (*it).c_str());
			}
			for(it = numbers.begin(); it != numbers.end(); ++it) {
				fprintf(fp_d, "\t%s", (*it).c_str());
			}

			fprintf(fp_d, "\n");
		}
	}
	
	cerr << "### main_loop: starting time loop ###" << endl;
	
	if(Offset == 0) {
		theCup->takeMolSnapshot(fp_a, 0.0);
	}
	
	
	for(it=0; it<Length; ++it, ++i_out, ++i_report)
	{

		// advance the simulation one step
		theCup->doOneStep(it);
		
		// output of molecule positions ..
		if(i_out >= Out_var)
		{
			theCup->takeMolSnapshot(fp_a, (it+1+Offset)*Time_Step);
		
			// .. and densities, numbers
			if(strlen(outputFile) > 0) 
			{
				fprintf(fp_d, "%lf", (it+1+Offset)*Time_Step);
				
				// ask Cup for densities
				std::list<std::string>::iterator it;
				for(it = densities.begin(); it != densities.end(); ++it) 
				{
					double dens = theCup->getDensity((*it));
					fprintf(fp_d, "\t%.3e", dens);
				}
				
				// ask cup for particle numbers
				for(it = numbers.begin(); it != numbers.end(); ++it) 
				{
					long int anz = theCup->countMolecules((*it));
					fprintf(fp_d, "\t%ld", anz);
				}
				fprintf(fp_d, "\n");
			}
			
			
			// reset output counter
			i_out = 0;
			++outCounter;
		}
		
	}
	
	cerr << "### main_loop: simulation done ########" << endl;
	cerr << "#######################################" << endl;
	
	if(strlen(outputFile) > 0) {
		fclose(fp_d);
	}
	
	if(!noDump) {
		theCup->dumpMolecules();
	}
}


// tear down the whole system again
void finalize(Cup *theCup)
{
	cerr << "### finalize: tear down cup with all its content" << endl;
	delete theCup;
	theCup = NULL;
}


int main(int argc, char **argv)
{
	Cup *theCup = NULL;
	theCup = new Cup();  
	FILE *fp_a;
	double Time_Step;
	long Length, Out_var, Offset;
	bool noDump = false;
	char commandline[2047];
	char outputFile[2048];
	
	// save names of reservoir and particle labels for output
	list<std::string> densities;
	list<std::string> numbers;
	// list of external forces
	list<externalForce> felder;
	
	if(argc != 2) {
		cerr << "Brownmove 1.0" << endl;
		cerr << "   compiled" <<
		#include "compileInfo.h"
		<< endl;
		cerr << "\n  Usage:  " << argv[0] << "  <setup file>" << endl;
		return(-1);
	}
	
	initialize();
	
	Offset = 0;	// this is an otional parameter in the setup
	init_cup(theCup, argv[1], commandline, Time_Step, Length, Out_var, Offset, noDump, outputFile, densities, numbers, felder);

	// must be opened here to survive the initialization
	if(!(fp_a = popen(commandline, "w")))
	{
		std::cerr << "main::main: could not open pipe to analyzer >>" << commandline << "<<" << std::endl;
		exit(-2);
	}
	std::cerr << "main::init_cup: using analyzer script >>" << commandline << "<<" << std::endl;

	// actually run the simulation
	main_loop(theCup, fp_a, Time_Step, Length, Out_var, Offset, noDump, outputFile, densities, numbers);
	
	// clean up
	pclose(fp_a);
	finalize(theCup);
	
	return (0);
}
